package net.sf.sido.pojo.tool;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import net.sf.sido.*;
import net.sf.sido.array.Array;
import net.sf.sido.array.IndexedArray;
import net.sf.sido.constraint.EnumConstraint;
import net.sf.sido.factory.UniversalDataFactoryReader;
import net.sf.sido.pojo.*;
import net.sf.sido.pojo.tool.model.JClass;
import net.sf.sido.pojo.tool.model.JConstructor;
import net.sf.sido.support.DOFunctions;
import net.sf.sido.support.DataFactories;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.StringUtils;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

public class Tool {

    private Set<File> sourceFiles;
    private File outputDirectory;
    private File servicesDirectory;
    private String outputPackage;

    public Set<File> getSourceFiles() {
        return sourceFiles;
    }

    public void setSourceFiles(Set<File> sourceFiles) {
        this.sourceFiles = sourceFiles;
    }

    public File getOutputDirectory() {
        return outputDirectory;
    }

    public void setOutputDirectory(File outputDirectory) {
        this.outputDirectory = outputDirectory;
    }

    public String getOutputPackage() {
        return outputPackage;
    }

    public void setOutputPackage(String outputPackage) {
        this.outputPackage = outputPackage;
    }

    public File getServicesDirectory() {
        return servicesDirectory;
    }

    public void setServicesDirectory(File servicesDirectory) {
        this.servicesDirectory = servicesDirectory;
    }

    public void generate(final ToolListener listener) throws IOException {
        // Log
        listener.log("Current class loader is: %s", Thread.currentThread().getContextClassLoader());

        // Creates a factory for all the SiDOL files
        listener.log("Creating the DataFactory...");
        DataFactory dataFactory = DataFactories.typedFactory();

        // Converts files to URLs
        Collection<URL> urls = Collections2.transform(sourceFiles,
                new Function<File, URL>() {
                    @Override
                    public URL apply(File file) {
                        try {
                            return file.toURI().toURL();
                        } catch (MalformedURLException e) {
                            throw new IllegalStateException("Cannot get file URL for " + file, e);
                        }
                    }
                });

        // Logger wrapper
        DOFactoryLogger factoryLogger = new DOFactoryLogger() {
            @Override
            public void log(String pattern, Object... parameters) {
                listener.log(pattern, parameters);
            }
        };

        // Loads and discovers other schemas
        listener.log("Discovering other schemas...");
        DOFactory doFactory = new DOFactory(dataFactory, true, factoryLogger);

        // Logs the DataSchemas
        listener.log("Imported schemas are:");
        Collection<DataSchema> schemas = doFactory.getDataFactory().getSchemas();
        for (DataSchema schema : schemas) {
            listener.log(" * %s", schema.getName());
        }

        // Logs the DO schemas
        listener.log("Imported DO schemas are:");
        Collection<DOSchema> doSchemas = doFactory.getSchemas();
        for (DOSchema doSchema : doSchemas) {
            listener.log(" * %s -> %s", doSchema.getDataSchema().getName(), doSchema.getClass().getName());
        }

        // Loads specified files
        UniversalDataFactoryReader reader = new UniversalDataFactoryReader();
        Collection<DataSchema> dataSchemas = reader.read(urls, dataFactory);

        // All classes to generate
        Set<JClass> jClasses = new HashSet<JClass>();

        // Generates all schemas and types
        Collection<JClass> schemaClasses = generate(listener, jClasses, dataFactory);

        // Writes all classes
        writeClasses(listener, jClasses);

        // Registers the schemas
        File serviceFile = new File(servicesDirectory, "META-INF/sido/sido.pojo");
        serviceFile.getParentFile().mkdirs();
        PrintWriter writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(serviceFile), "UTF-8"));
        try {
            for (JClass schemaClass : schemaClasses) {
                writer.format("%s = %s%n", schemaClass.getInfo(), schemaClass.getQualifiedName());
            }
        } finally {
            writer.close();
        }

        // Copies all schemas
        Iterator<DataSchema> dataSchemaIterator = dataSchemas.iterator();
        for (URL url : urls) {
            DataSchema dataSchema = dataSchemaIterator.next();
            // Creates the SiDOL file
            File sidolFile = new File(servicesDirectory, String.format("META-INF/sido/%s.sidol", dataSchema.getName()));
            FileUtils.copyURLToFile(url, sidolFile);
        }
    }

    private void writeClasses(ToolListener listener, Set<JClass> jClasses) throws IOException {
        for (JClass jClass : jClasses) {
            writeClass(listener, jClass);
        }
    }

    private void writeClass(ToolListener listener, JClass jClass) throws IOException {
        String className = jClass.getName();
        String fileName = className + ".java";
        String packagePath = StringUtils.replace(outputPackage, ".", "/");
        File packageFile = new File(outputDirectory, packagePath);
        if (!packageFile.exists()) {
            packageFile.mkdirs();
        }
        File file = new File(packageFile, fileName);
        PrintWriter writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF-8"));
        try {
            jClass.write(writer);
        } finally {
            writer.close();
        }
    }

    private Collection<JClass> generate(ToolListener listener, Set<JClass> jClasses, DataFactory dataFactory) {
        Collection<JClass> schemaClasses = new ArrayList<JClass>();
        // Gets all schemas
        Collection<DataSchema> schemas = dataFactory.getSchemas();
        for (DataSchema schema : schemas) {
            JClass schemaClass = generateSchema(listener, jClasses, schema);
            schemaClasses.add(schemaClass);
        }
        // OK
        return schemaClasses;
    }

    private JClass generateSchema(ToolListener listener, Set<JClass> jClasses, DataSchema schema) {
        String name = schema.getName();
        String className = name + "Schema";
        log(listener, "Generating schema %s", name);
        // Class
        JClass j = new JClass(outputPackage, className);
        // TODO Header
        // Imports
        j.addImport(DataObject.class);
        j.addImport(DOFactory.class);
        j.addImport(DOSchema.class);
        // Parent
        j.setParent("DOSchema");
        // Doc
        j.addDoc("Schema for %s", name);
        j.setInfo(name);
        // Constructor
        JConstructor jConstructor = j.addConstructor()
                .addParam("DOFactory", "doFactory")
                .addContent("super(doFactory, \"%s\");", name);
        // All types
        for (DataType type : schema.getTypes()) {
            if (!type.isAbstractType()) {
                String typeName = type.getName();
                // Registration of the type in the constructor
                jConstructor.addContent("registerClass(\"%1$s::%2$s\", %1$s%2$s.class);", name, typeName);
                // Simple creation
                j.addMethod("create" + typeName, name + typeName)
                        .addContent("return new %1$s%2$s(this, newInstance(\"%1$s::%2$s\"));", name, typeName);
                // Creation from JSON
                j.addMethod("create" + typeName, name + typeName)
                        .addParam("String", "json")
                        .addContent("DataObject o = read(\"%s::%s\", json);", name, typeName)
                        .addContent("return create(o);");
            }
        }
        // OK
        jClasses.add(j);
        // Generates the types
        for (DataType type : schema.getTypes()) {
            generateType(listener, jClasses, type);
        }
        // OK
        return j;
    }

    protected void generateType(ToolListener listener, Set<JClass> jClasses, DataType type) {
        String schema = type.getSchema().getName();
        String name = type.getName();
        String className = schema + name;
        log(listener, "Generating type %s", className);
        // Class
        JClass j = new JClass(outputPackage, className);
        // TODO Header
        // Imports
        j.addImport("net.sf.sido.DataObject");
        // Type class
        String parentTypeName;
        // Inheritance
        DataType parentType = type.getParentType();
        String superSchema = "schema";
        if (parentType != null) {
            parentTypeName = String.format("%s%s", parentType.getSchema().getName(), parentType.getName());
            // Gets the schema of the parent
            DataSchema parentSchema = parentType.getSchema();
            // If another schema
            if (!parentSchema.getName().equals(schema)) {
                superSchema = String.format("schema.getSchema(%sSchema.class)", parentSchema.getName());
                // FIXME Missing import for the parent schema
            }
        } else {
            // Default parent
            parentTypeName = "DOObject";
            // Imports the parent
            j.addImport("net.sf.sido.pojo.DOObject");
        }
        // Abstract class
        if (type.isAbstractType()) {
            j.setAbstractClass(true);
        }
        // Parent type
        j.setParent(parentTypeName);
        // Class doc
        j.addDoc("POJO class for %s::%s", schema, name);
        // Constructor
        j.addConstructor()
                .addParam(schema + "Schema", "schema")
                .addParam("DataObject", "o")
                .addContent("super(%s,o);", superSchema);
        // Support for the contextual factory
        if (type.isAbstractType()) {
            j.addImport(DOSchema.class);
            j.addConstructor()
                    .addParam("Class<? extends DOSchema>", "schemaClass")
                    .addParam("String", "typeName")
                    .addContent("super(schemaClass, typeName);");
        } else {
            j.addConstructor()
                    .addContent("super(%s.class, \"%s::%s\");", schema + "Schema", schema, name);
        }
        // Schema
        j.addMethod(String.format("get%sSchema", schema), schema + "Schema")
                .addContent("return getSchema(%sSchema.class);", schema);
        // Type
        j.addMethod("getQualifiedTypeName", "String")
                .setStatic()
                .addContent("return \"%s::%s\";", schema, name);
        // All properties
        Collection<Property<?>> properties = type.getProperties();
        for (Property<?> property : properties) {
            generateProperty(listener, j, property);
        }
        // End of the type class
        jClasses.add(j);
    }

    protected <T> void generateProperty(ToolListener listener, JClass j, Property<T> property) {
        String propertyName = property.getName();
        if (property.isSimple()) {
            if (property.isArray()) {
                // Array + Simple
                Class<?> pType = property.getMetaInfo().getType(PropertyMetaInfos.ELEMENT_TYPE);
                String pTypeName = pType.getSimpleName();
                String arrayTypeName = String.format("Array<%s>", pTypeName);
                // Imports the type
                j.addImport(pType);
                j.addImport(Array.class);
                // Getter
                j.addMethod(getGetMethodName(propertyName), arrayTypeName)
                        .addContent("return o.getArray(\"%s\");", propertyName);
                // Setter
                j.addMethod(getSetMethodName(propertyName))
                        .addParam(arrayTypeName, "values")
                        .addContent("o.setArray(\"%s\", values);", propertyName);
                // Setter for a collection
                j.addImport(Collection.class);
                j.addMethod(getSetMethodName(propertyName))
                        .addParam(String.format("Collection<%s>", pTypeName), "values")
                        .addContent("o.setArray(\"%s\", values);", propertyName);
                // Add
                j.addMethod(getAddMethodName(propertyName))
                        .addParam(String.format("%s...", pTypeName), "values")
                        .addContent("o.add(\"%s\", values);", propertyName);
                // Index of
                j.addMethod(getIndexOfMethodName(propertyName), "int")
                        .addParam(pTypeName, "value")
                        .addContent("Array<%s> array = o.getArray(\"%s\");", pTypeName, propertyName)
                        .addContent("return array != null ? array.indexOf(value) : -1;");
                // Delete
                j.addMethod(getDeleteMethodName(propertyName))
                        .addParam("int...", "indexes")
                        .addContent("Array<%s> array = o.getArray(\"%s\");", pTypeName, propertyName)
                        .addContent("if (array != null) {")
                        .addContent("\tArray<%s> result = array.delete(indexes);", pTypeName)
                        .addContent("\tif (result != array) {")
                        .addContent("\t\to.setArray(\"%s\", result);", propertyName)
                        .addContent("\t}")
                        .addContent("}");
                // Remove
                j.addMethod(getRemoveMethodName(propertyName))
                        .addParam(pTypeName, "value")
                        .addContent("Array<%s> array = o.getArray(\"%s\");", pTypeName, propertyName)
                        .addContent("if (array != null) {")
                        .addContent("\tArray<%s> result = array.remove(value);", pTypeName)
                        .addContent("\tif (result != array) {")
                        .addContent("\t\to.setArray(\"%s\", result);", propertyName)
                        .addContent("\t}")
                        .addContent("}");
            } else {
                // Simple
                Class<? super T> pType = property.getType();
                String pTypeName;
                // Enum constraint?
                Class<Enum<?>> enumType = getEnumType(property);
                if (enumType != null) {
                    // Imports the enum
                    j.addImport(enumType);
                    // Type name
                    pTypeName = enumType.getSimpleName();
                    // Getter
                    j.addMethod(getGetMethodName(propertyName), pTypeName)
                            .addContent("return o.getEnum(\"%s\", %s.class);", propertyName, pTypeName);
                    // Setter
                    j.addMethod(getSetMethodName(propertyName))
                            .addParam(pTypeName, "value")
                            .addContent("o.setEnum(\"%s\", value);", propertyName);
                }
                // Converts to primitive type is not nullable
                else {
                    String accessor = pType.getSimpleName();
                    if (!property.isNullable()) {
                        // Gets the primitive type
                        Class<?> primitiveType = getPrimitiveType(pType);
                        if (primitiveType != null) {
                            pTypeName = primitiveType.getSimpleName();
                        } else {
                            pTypeName = accessor;
                            // Imports this type (only if not primitive)
                            j.addImport(pType);
                        }
                    } else {
                        pTypeName = accessor;
                        // Imports this type (only if not primitive)
                        j.addImport(pType);
                    }
                    // Getter
                    j.addMethod(getGetMethodName(propertyName), pTypeName)
                            .addContent("return o.get%s(\"%s\");", accessor, propertyName);
                    // Setter
                    j.addMethod(getSetMethodName(propertyName))
                            .addParam(pTypeName, "value")
                            .addContent("o.set%s(\"%s\", value);", accessor, propertyName);
                }
            }
        } else {
            if (property.isArray()) {
                // Array + DO
                DataType elementType = property.getMetaInfo().getObject(PropertyMetaInfos.ELEMENT_TYPE, null);
                // Array of anonymous
                if (elementType == null) {
                    generateAnonymousArrayProperty(j, propertyName);
                }
                // Indexed array
                else if (property.isIndexedArray()) {
                    generatedIndexedArrayProperty(j, property, propertyName, elementType);

                }
                // Array of DO
                else {
                    generateArrayDOProperty(j, propertyName, elementType);
                }
            } else {
                // DO
                DataType doType = property.getMetaInfo().getObject(PropertyMetaInfos.DO_TYPE, null);
                // Anonymous
                if (doType == null) {
                    generateAnonymousProperty(j, propertyName);
                }
                // DO type
                else {
                    generateDOProperty(j, propertyName, doType);
                }
            }
        }
    }

    /**
     * Indexed array.
     * <p/>
     * <pre>
     * private IndexedArray<CompleteCompany, String> companies;
     *
     * public IndexedArray<CompleteCompany, String> getCompanies() {
     * return (companies = getPojoIndexedArray(companies, o.<String>getIndexedArray("companies"), "id"));
     * }
     *
     * public void setCompanies(IndexedArray<CompleteCompany, String> values) {
     * o.setIndexedArray("companies", values != null ? values.transform(DOObject.getDataObjectFn(), DOFunctions.get("id")) : null);
     * companies = getPojoIndexedArray(companies, values, o.<String>getIndexedArray("companies"), "id");
     * }
     *
     * public void setCompanies(Collection<CompleteCompany> values) {
     * o.setIndexedArray("companies", values != null ? Collections2.transform(values, DOObject.getDataObjectFn()) : null);
     * companies = getPojoIndexedArray(companies, values, o.<String>getIndexedArray("companies"), "id");
     * }
     *
     * public void addCompanies(CompleteCompany... values) {
     * o.add("companies", DOUtils.transform(values));
     * companies = getPojoIndexedArray(companies, Arrays.asList(values), o.<String>getIndexedArray("companies"), "id");
     * }
     *
     * public CompleteCompany getCompanyByIndex(String key) {
     * companies = getPojoIndexedArray(companies, o.<String>getIndexedArray("companies"), "id");
     * return companies != null ? companies.getByIndex(key) : null;
     * }
     * </pre>
     */
    private <T> void generatedIndexedArrayProperty(JClass j, Property<T> property, String propertyName, DataType elementType) {
        // Gets the index information
        String indexPropertyName = property.getMetaInfo().getString(PropertyMetaInfos.ARRAY_INDEX_PROPERTY, null);
        // Gets the property on the element type
        Property<?> indexProperty = elementType.getProperty(indexPropertyName);
        Class<?> indexType = indexProperty.getType();
        // Element class
        JClass elementClass = getClass(elementType);
        j.addImport(elementClass);
        // Indexed array type
        String indexTypeSimpleName = indexType.getSimpleName();
        String indexedArrayType = String.format("IndexedArray<%s,%s>", elementClass.getName(), indexTypeSimpleName);
        // Imports
        j.addImport(indexType);
        j.addImport(elementClass);
        j.addImport(Array.class);
        j.addImport(Arrays.class);
        j.addImport(DOObject.class);
        j.addImport(DOFunctions.class);
        j.addImport(Collection.class);
        j.addImport(IndexedArray.class);
        j.addImport(DOUtils.class);
        j.addImport(Collections2.class);
        // Field
        j.addField(indexedArrayType, propertyName);
        // Getter
        j.addMethod(getGetMethodName(propertyName), indexedArrayType)
                .addContent("return (%2$s = getPojoIndexedArray(%2$s, o.<%1$s>getIndexedArray(\"%2$s\"), \"%3$s\"));",
                        indexTypeSimpleName, propertyName, indexPropertyName);
        // Setter from an array
        j.addMethod(getSetMethodName(propertyName))
                .addParam(indexedArrayType, "values")
                .addContent("o.setIndexedArray(\"%1$s\", values != null ? values.transform(DOObject.getDataObjectFn(), DOFunctions.get(\"%2$s\")) : null);", propertyName, indexPropertyName)
                .addContent("%2$s = getPojoIndexedArray(%2$s, values, o.<%1$s>getIndexedArray(\"%2$s\"), \"%3$s\");",
                        indexTypeSimpleName, propertyName, indexPropertyName);
        // Setter from a collection
        j.addMethod(getSetMethodName(propertyName))
                .addParam(String.format("Collection<%s>", elementClass.getName()), "values")
                .addContent("o.setIndexedArray(\"%s\", values != null ? Collections2.transform(values, DOObject.getDataObjectFn()) : null);",
                        propertyName)
                .addContent("%2$s = getPojoIndexedArray(%2$s, values, o.<%1$s>getIndexedArray(\"%2$s\"), \"%3$s\");",
                        indexTypeSimpleName, propertyName, indexPropertyName);
        // Adding new elements
        j.addMethod(getAddMethodName(propertyName))
                .addParam(elementClass.getName() + "...", "values")
                .addContent("o.add(\"%s\", DOUtils.transform(values));", propertyName)
                .addContent("%2$s = getPojoIndexedArray(%2$s, Arrays.asList(values), o.<%1$s>getIndexedArray(\"%2$s\"), \"%3$s\");",
                        indexTypeSimpleName, propertyName, indexPropertyName);
        // Getting by index
        j.addMethod(getByIndexMethodName(propertyName), elementClass.getName())
                .addParam(indexTypeSimpleName, "key")
                .addContent("%2$s = getPojoIndexedArray(%2$s, o.<%1$s>getIndexedArray(\"%2$s\"), \"%3$s\");",
                        indexTypeSimpleName, propertyName, indexPropertyName)
                .addContent("return %1$s != null ? %1$s.getByIndex(key) : null;", propertyName);
        // Index of
        j.addMethod(getIndexOfMethodName(propertyName), "int")
                .addParam(elementClass.getName(), "value")
                .addContent("Array<%s> array = o.getArray(\"%s\");", elementClass.getName(), propertyName)
                .addContent("return array != null ? array.indexOf(value) : -1;");
        // Delete
        j.addMethod(getDeleteMethodName(propertyName))
                .addParam("int...", "indexes")
                .addContent("Array<%s> array = o.getArray(\"%s\");", elementClass.getName(), propertyName)
                .addContent("if (array != null) {")
                .addContent("\tArray<%s> result = array.delete(indexes);", elementClass.getName())
                .addContent("\tif (result != array) {")
                .addContent("\t\to.setArray(\"%s\", result);", propertyName)
                .addContent("\t}")
                .addContent("}");
        // Remove
        j.addMethod(getRemoveMethodName(propertyName))
                .addParam(elementClass.getName(), "value")
                .addContent("Array<%s> array = o.getArray(\"%s\");", elementClass.getName(), propertyName)
                .addContent("if (array != null) {")
                .addContent("\tArray<%s> result = array.remove(value);", elementClass.getName())
                .addContent("\tif (result != array) {")
                .addContent("\t\to.setArray(\"%s\", result);", propertyName)
                .addContent("\t}")
                .addContent("}");
    }

    /**
     * Anonymous property.
     * <p/>
     * <pre>
     * private DOObject data;
     *
     * public DOObject getData() {
     * return data != null ? data : (data = getSchema().create(o.getDataObject("data")));
     * }
     *
     * public void setData(DOObject value) {
     * o.setDataObject("data", value != null ? value.getDataObject() : null);
     * this.data = value;
     * }
     * </pre>
     */
    private void generateAnonymousProperty(JClass j, String propertyName) {
        j.addImport(DOObject.class);
        // Field
        j.addField("DOObject", propertyName);
        // Getter
        j.addMethod(getGetMethodName(propertyName), "DOObject")
                .addContent("return %1$s != null ? %1$s : (%1$s = getSchema().create(o.getDataObject(\"%1$s\")));", propertyName);
        // Setter
        j.addMethod(getSetMethodName(propertyName))
                .addParam("DOObject", "value")
                .addContent("o.setDataObject(\"%s\", value != null ? value.getDataObject() : null);", propertyName)
                .addContent("this.%s = value;", propertyName);
    }

    /**
     * Anonymous array property.
     * <p/>
     * <pre>
     *
     * private Array<DOObject> params;
     *
     * public Array<DOObject> getParams() {
     * return (params = getPojoArray(params, o.getDOArray("params")));
     * }
     *
     * public void setParams(Array<? extends DOObject> values) {
     * o.setArray("params", values != null ? values.transform(DOObject.getDataObjectFn()) : null);
     * params = getPojoArray(params, values, o.getDOArray("params"));
     * }
     *
     * public void setParams(Collection<? extends DOObject> values) {
     * o.setArray("params", values != null ? Collections2.transform(values, DOObject.getDataObjectFn()) : null);
     * params = getPojoArray(params, values, o.getDOArray("params"));
     * }
     *
     * public void addParams(DOObject... values) {
     * o.add("params", DOUtils.transform(values));
     * params = getPojoArray(params, Arrays.asList(values), o.getDOArray("params"));
     * }
     *
     * </pre>
     */
    private void generateAnonymousArrayProperty(JClass j, String propertyName) {
        j.addImport(Array.class);
        j.addImport(DOUtils.class);
        j.addImport(DOObject.class);
        // Field
        j.addField("Array<DOObject>", propertyName);
        // Getter
        j.addMethod(getGetMethodName(propertyName), "Array<DOObject>")
                .addContent("return (%1$s = getPojoArray(%1$s, o.getDOArray(\"%1$s\")));", propertyName);
        // Setter
        j.addMethod(getSetMethodName(propertyName))
                .addParam("Array<? extends DOObject>", "values")
                .addContent("o.setArray(\"%1$s\", values != null ? values.transform(DOObject.getDataObjectFn()) : null);", propertyName)
                .addContent("%1$s = getPojoArray(%1$s, values, o.getDOArray(\"%1$s\"));", propertyName);
        // Setter from collection
        j.addImport(Collection.class);
        j.addImport(Collections2.class);
        j.addMethod(getSetMethodName(propertyName))
                .addParam("Collection<? extends DOObject>", "values")
                .addContent("o.setArray(\"%s\", values != null ? Collections2.transform(values, DOObject.getDataObjectFn()) : null);", propertyName)
                .addContent("%1$s = getPojoArray(%1$s, values, o.getDOArray(\"%1$s\"));", propertyName);
        // Add
        j.addImport(Arrays.class);
        j.addMethod(getAddMethodName(propertyName))
                .addParam("DOObject...", "values")
                .addContent("o.add(\"%s\", DOUtils.transform(values));", propertyName)
                .addContent("%1$s = getPojoArray(%1$s, Arrays.asList(values), o.getDOArray(\"%1$s\"));", propertyName);
    }

    /**
     * DO property.
     * <p/>
     * <pre>
     * private CompleteAddress address;
     *
     * public CompleteAddress getAddress() {
     * return address != null ? address : (address = getSchema().create(o.getDataObject("address")));
     * }
     *
     * public void setAddress(CompleteAddress value) {
     * o.setDataObject("address", value != null ? value.getDataObject() : null);
     * this.address = address;
     * }
     * </pre>
     */
    protected void generateDOProperty(JClass j, String propertyName, DataType doType) {
        JClass doTypeClass = getClass(doType);
        // Field
        j.addField(doTypeClass, propertyName);
        // Getter
        j.addMethod(getGetMethodName(propertyName), doTypeClass)
                .addContent("return %1$s != null ? %1$s : (%1$s = getSchema().create(o.getDataObject(\"%1$s\")));", propertyName);
        // Setter
        j.addMethod(getSetMethodName(propertyName))
                .addParam(doTypeClass, "value")
                .addContent("o.setDataObject(\"%s\", value != null ? value.getDataObject() : null);", propertyName)
                .addContent("this.%s = value;", propertyName);
    }

    /**
     * DO array.
     * <pre>
     * private Array<CompletePerson> contacts;
     *
     * public Array<CompletePerson> getContacts() {
     * return (contacts = getPojoArray(contacts, o.getDOArray("contacts")));
     * }
     *
     * public void setContacts(Array<CompletePerson> values) {
     * o.setArray("contacts", values.transform(DOObject.getDataObjectFn()));
     * contacts = getPojoArray(contacts, values, o.getDOArray("contacts"));
     * }
     *
     * public void setContacts(Collection<CompletePerson> values) {
     * o.setArray("contacts", Collections2.transform(values, DOObject.getDataObjectFn()));
     * contacts = getPojoArray(contacts, values, o.getDOArray("contacts"));
     * }
     *
     * public void addContacts(CompletePerson... values) {
     * o.add("contacts", DOUtils.transform(values));
     * contacts = getPojoArray(contacts, Arrays.asList(values), o.getDOArray("contacts"));
     * }
     * </pre>
     */
    protected void generateArrayDOProperty(JClass j, String propertyName, DataType elementType) {
        j.addImport(Array.class);
        j.addImport(DOUtils.class);
        j.addImport(DOObject.class);
        j.addImport(Arrays.class);
        JClass doClass = getClass(elementType);
        j.addImport(doClass);
        String arrayClassName = String.format("Array<%s>", doClass.getName());
        // Field
        j.addField(arrayClassName, propertyName);
        // Getter
        j.addMethod(getGetMethodName(propertyName), arrayClassName)
                .addContent("return (%1$s = getPojoArray(%1$s, o.getDOArray(\"%1$s\")));", propertyName);
        // Setter
        j.addMethod(getSetMethodName(propertyName))
                .addParam(arrayClassName, "values")
                .addContent("o.setArray(\"%1$s\", values.transform(DOObject.getDataObjectFn()));", propertyName)
                .addContent("%1$s = getPojoArray(%1$s, values, o.getDOArray(\"%1$s\"));", propertyName);
        // Setter from collection
        j.addImport(Collection.class);
        j.addImport(Collections2.class);
        j.addMethod(getSetMethodName(propertyName))
                .addParam(String.format("Collection<%s>", doClass.getName()), "values")
                .addContent("o.setArray(\"%1$s\", Collections2.transform(values, DOObject.getDataObjectFn()));", propertyName)
                .addContent("%1$s = getPojoArray(%1$s, values, o.getDOArray(\"%1$s\"));", propertyName);
        // Add
        j.addMethod(getAddMethodName(propertyName))
                .addParam(String.format("%s...", doClass.getName()), "values")
                .addContent("o.add(\"%1$s\", DOUtils.transform(values));", propertyName)
                .addContent("%1$s = getPojoArray(%1$s, Arrays.asList(values), o.getDOArray(\"%1$s\"));", propertyName);
        // Index of
        j.addMethod(getIndexOfMethodName(propertyName), "int")
                .addParam(doClass.getName(), "value")
                .addContent("Array<%s> array = o.getArray(\"%s\");", doClass.getName(), propertyName)
                .addContent("return array != null ? array.indexOf(value) : -1;");
        // Delete
        j.addMethod(getDeleteMethodName(propertyName))
                .addParam("int...", "indexes")
                .addContent("Array<%s> array = o.getArray(\"%s\");", doClass.getName(), propertyName)
                .addContent("if (array != null) {")
                .addContent("\tArray<%s> result = array.delete(indexes);", doClass.getName())
                .addContent("\tif (result != array) {")
                .addContent("\t\to.setArray(\"%s\", result);", propertyName)
                .addContent("\t}")
                .addContent("}");
        // Remove
        j.addMethod(getRemoveMethodName(propertyName))
                .addParam(doClass.getName(), "value")
                .addContent("Array<%s> array = o.getArray(\"%s\");", doClass.getName(), propertyName)
                .addContent("if (array != null) {")
                .addContent("\tArray<%s> result = array.remove(value);", doClass.getName())
                .addContent("\tif (result != array) {")
                .addContent("\t\to.setArray(\"%s\", result);", propertyName)
                .addContent("\t}")
                .addContent("}");
    }

    private Class<Enum<?>> getEnumType(Property<?> property) {
        Class<?> propertyType = property.getType();
        if (String.class.isAssignableFrom(propertyType)) {
            Predicate<?> constraint = property.getConstraint();
            if (constraint != null && constraint instanceof EnumConstraint) {
                return ((EnumConstraint) constraint).getEnumClass();
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    private Class<?> getPrimitiveType(Class<?> type) {
        return ClassUtils.wrapperToPrimitive(type);
    }

    private JClass getClass(DataType type) {
        // FIXME Get the package for other schemas
        return new JClass(outputPackage, type.getSchema().getName() + type.getName());
    }

    private String getSetMethodName(String name) {
        return getAccessorName("set", name);
    }

    private String getAccessorName(String prefix, String name) {
        return String.format("%s%s", prefix, StringUtils.capitalize(name));
    }

    private String getGetMethodName(String name) {
        return getAccessorName("get", name);
    }

    private String getAddMethodName(String name) {
        return getAccessorName("add", name);
    }

    private String getByIndexMethodName(String name) {
        return getAccessorName("get", name) + "ByIndex";
    }

    private String getIndexOfMethodName(String name) {
        return getAccessorName("indexOf", name);
    }

    private String getDeleteMethodName(String name) {
        return getAccessorName("delete", name);
    }

    private String getRemoveMethodName(String name) {
        return getAccessorName("remove", name);
    }

    private void log(ToolListener listener, String pattern, Object... params) {
        listener.log(pattern, params);
    }
}
